module MonoidFoldable where

import Data.Foldable
import Data.Semigroup
import Data.Monoid hiding ((<>))

type Position = (Double, Double)
type Length   = Double
type Area     = Double

data Object
    = Rectangle {centre :: Position, width :: Length, height :: Length}
    | Circle    {centre :: Position, radius :: Length}


circumference :: Object -> Length
circumference (Rectangle {width = w, height = h}) = 2 * (w + h)
circumference (Circle {radius = r}) = 2 * pi * r

area :: Object -> Area
area (Rectangle {width = w, height = h}) = w * h
area (Circle {radius = r}) = 4 * pi * r / 3

data Drawing a
    = Element a
    | Group  [Drawing a]

data Statistics = 
    Statistics {
        avgArea          :: Area,
        avgCircumference :: Length,
        maxArea          :: Area,
        maxCircumference :: Length
    }
    deriving Show

data AccumStats =
    AccumStats {
        asCount            :: Int,
        asSumArea          :: Area,
        asSumCircumference :: Length,
        asMaxArea          :: Area,
        asMaxCircumference :: Length
    }

statistics :: Drawing Object -> Statistics
statistics d
    = Statistics {
         avgArea          = asSumArea as / fromIntegral (asCount as),
         avgCircumference = asSumCircumference as / fromIntegral (asCount as),
         maxArea          = asMaxArea as,
         maxCircumference = asMaxCircumference as

      }
    where
        as = statsAux d

        statsAux (Element o) =
            AccumStats {
                asCount            = 1,
                asSumArea          = a, 
                asSumCircumference = c,
                asMaxArea          = a,
                asMaxCircumference = c
            }
            where
                a = area o
                c = circumference o
        statsAux (Group ds) = statsList dsa0 ds
            where
                dsa0 = AccumStats {
                           asCount            = 0,
                           asSumArea          = 0, 
                           asSumCircumference = 0,
                           asMaxArea          = 0,
                           asMaxCircumference = 0
                       }

        statsList as []       = as
        statsList as (d : ds) = statsList as' ds
            where
                as1 = statsAux d

                as' = AccumStats {
                          asCount            = asCount as + asCount as1,
                          asSumArea          = asSumArea as
                                               + asSumArea as1,
                          asSumCircumference = asSumCircumference as
                                               + asSumCircumference as1,
                          asMaxArea          = asMaxArea as
                                               `max` asMaxArea as1,
                          asMaxCircumference = asMaxCircumference as
                                               `max` asMaxCircumference as1
                       }


instance Semigroup AccumStats where
    as1 <> as2 = AccumStats {
                     asCount            = asCount as1 + asCount as2,
                     asSumArea          = asSumArea as1
                                          + asSumArea as2,
                     asSumCircumference = asSumCircumference as1
                                          + asSumCircumference as2,
                     asMaxArea          = asMaxArea as1
                                          `max` asMaxArea as2,
                     asMaxCircumference = asMaxCircumference as1
                                          `max` asMaxCircumference as2
                 }



instance Monoid AccumStats where
    mempty = AccumStats {
                 asCount            = 0,
                 asSumArea          = 0, 
                 asSumCircumference = 0,
                 asMaxArea          = 0,
                 asMaxCircumference = 0
             }

    mappend = (<>)

instance Foldable Drawing where
    foldMap f (Element e) = f e
    foldMap f (Group ds)  = foldMap (foldMap f) ds

statistics1 :: Drawing Object -> Statistics
statistics1 d =
    Statistics {
       avgArea          = asSumArea as / fromIntegral (asCount as),
       avgCircumference = asSumCircumference as / fromIntegral (asCount as),
       maxArea          = asMaxArea as,
       maxCircumference = asMaxCircumference as
    }
    where
        as = foldMap asObject d

asObject :: Object -> AccumStats
asObject o =
    AccumStats {
        asCount            = 1,
        asSumArea          = a, 
        asSumCircumference = c,
        asMaxArea          = a,
        asMaxCircumference = c
    }
    where
        a = area o
        c = circumference o


testDrawing :: Drawing Object
testDrawing =
    Group [
        Group [
            Element (Rectangle {centre = (10,10), width = 5, height = 7}),
            Element (Circle {centre = (11,11), radius = 5})
        ],
        Element (Rectangle {centre = (5,5), width = 10, height = 5})
    ]


-- Take 2: foldabe, simple monoid
-- Take 3: foldable, Sum, Max, use NewType for Length, Area to make bounded.
